---
icon: IconDirectory
title: 'layouts'
head.title: Layouts directory
---

# Layouts directory

Nuxt provides a customizable layouts framework you can use throughout your application, ideal for extracting common UI or code patterns into reusable layout components.

Page layouts are placed in the `layouts/` directory and will be automatically loaded via asynchronous import when used. If you create a `layouts/default.vue` this will be used for all pages in your app. Other layouts are used by setting a `layout` property as part of your component options.

If you have only single layout for application, you can alternatively use [app.vue](/docs/directory-structure/app).

## Example: a custom layout

```bash
-| layouts/
---| custom.vue
```

In your layout files, you'll need to use `<slot />` to define where the page content of your layout will be loaded. For example:

```vue
<template>
  <div>
    Some shared layout content:
    <slot />
  </div>
</template>
```

Given the example above, you can use a custom layout like this:

```vue
<script>
export default {
  layout: "custom",
};
</script>
```

## Example: using with slots

You can also take full control (for example, with slots) by using the `<NuxtLayout>` component (which is globally available throughout your application) and set `layout: false` in your component options.

```vue
<template>
  <NuxtLayout name="custom">
    <template #header> Some header template content. </template>

    The rest of the page
  </NuxtLayout>
</template>

<script>
export default {
  layout: false,
};
</script>
```

## Example: using with `<script setup>`

If you are utilizing Vue `<script setup>` [compile-time syntactic sugar](https://v3.vuejs.org/api/sfc-script-setup.html#sfc-script-setup), you can use a secondary `<script>` tag to set `layout` options as needed.

::alert{type=info}
Learn more about [`<script setup>` and `<script>` tags co-existing](https://v3.vuejs.org/api/sfc-script-setup.html#usage-alongside-normal-script) in the Vue docs.
::

Assuming this directory structure:

```bash
-| layouts/
---| custom.vue
-| pages/
---| my-page.vue
```

And this `custom.vue` layout:

```vue
<template>
  <div>
    Some shared layout content:
    <slot />
  </div>
</template>
```

You can set a page layout in `my-page.vue` — alongside the `<script setup>` tag — like this:

```vue
<script>
export default {
  layout: "custom",
};
</script>

<script setup>
// your setup script
</script>
```
